-- ****************************************************************************
--                                  namespace
-- ****************************************************************************

HotepMMX = {
  name = "HotepMMExtender",
  version = 1,
  savedVars = "HotepMMXVars",
  responding = false,
  asking = false,
  allguilds = {
    i = 0,
    num = false
  }
}



local HotepToolsLib = LibStub("HotepToolsLib")

local clone = HotepToolsLib.HotepCommonFuncs.clone
local explode = HotepToolsLib.HotepCommonFuncs.explode
local in_array = HotepToolsLib.HotepCommonFuncs.in_array
local array_key_exists = HotepToolsLib.HotepCommonFuncs.array_key_exists
local array_indexof = HotepToolsLib.HotepCommonFuncs.array_indexof
local array_without = HotepToolsLib.HotepCommonFuncs.array_without
local array_glob = HotepToolsLib.HotepCommonFuncs.array_glob
local uuid = HotepToolsLib.HotepCommonFuncs.uuid
local msgWithName = function (msg, color) HotepToolsLib.HotepCommonFuncs.msgWithName(msg, color, HotepMMX.name) end




local ONE_DAY = (3600 * 24)        -- one day in seconds
local TOO_OLD = (ONE_DAY * 60)


-- ****************************************************************************
--                                  templates
-- ****************************************************************************

local MMXSALE = {
  timestamp = 0,
  avgPrice = 0,
  numSales = 0,
  numItems = 0,
  numDays = 0,
}


local MMXITEM = {
  sales = {},
  avgPrice = 0,
  numSales = 0,
  numItems = 0,
  numDays = 0,
}





-- ****************************************************************************
--                                  saved vars
-- ****************************************************************************


local defaultVariables = {
  cleanup = {
    last = 0,            -- timestamp of last cleanup
    mule = 0,            -- mule # last cleaned up
  }
}


local savedVariables = {
  mules = {
    HotepMM01,
    HotepMM02,
    HotepMM03,
    HotepMM04,
    HotepMM05,
    HotepMM06,
    HotepMM07,
    HotepMM08,
    HotepMM09,
    HotepMM10,
    HotepMM11,
    HotepMM12,
    HotepMM13,
    HotepMM14,
    HotepMM15,
    HotepMM16,
  }
}

---
-- @local Cleanup @class MMXCleanup
local Cleanup


function savedVariables:Load()
  
  self.vars = ZO_SavedVars:NewAccountWide(HotepMMX.savedVars, HotepMMX.version, '', defaultVariables)
  
  Cleanup = self.vars.cleanup
  
end

function savedVariables:GetMuleIndex(itemLink)
  local icon, sellPrice, meetsUsageRequirement, 
        equipType, itemStyle = GetItemLinkInfo(itemLink)
  
  local itemType, specializedItemType = GetItemLinkItemType(itemLink)
  
  local x = sellPrice + ((itemType + 1) * 10)
  x = x + ((equipType + 3) * 25)
  x = x + ((itemStyle + 5) * 75)
  
  return ((x % 16) + 1)
end

function savedVariables:GetItemIndex(itemLink)
  local x = string.match(itemLink, '|H.-:item:(.-):')
  local y = MasterMerchant.makeIndexFromLink(itemLink)
  return zo_strformat("<<1>>$<<2>>", x, y)
end

---
-- @param itemLink @class string
-- @return @class MMXITEM
function savedVariables:GetItem(itemLink)
  local i = self:GetMuleIndex(itemLink)
  local itemIndex = self:GetItemIndex(itemLink)
  
  if (not self.mules[i].Vars.items[itemIndex]) then
    self.mules[i].Vars.items[itemIndex] = clone(MMXITEM)
  end
  
  return self.mules[i].Vars.items[itemIndex]
end

function savedVariables:GetMule(i)
  return self.mules[i].Vars.items
end

function savedVariables:DestroyItem(itemLink)
  local i = self:GetMuleIndex(itemLink)
  local itemIndex = self:GetItemIndex(itemLink)
  
  self.mules[i].Vars.items[itemIndex] = nil
end





-- ****************************************************************************
--                                  utilities
-- ****************************************************************************


local Timer = HotepToolsLib.HotepUtilities.Timer
local MailQueue = HotepToolsLib.HotepUtilities.MailQueue
local Iterator = HotepToolsLib.HotepUtilities.Iterator
local ChatQueue = HotepToolsLib.HotepUtilities.ChatQueue





-- ****************************************************************************
--                                  main code
-- ****************************************************************************


local MM_itemStats = MasterMerchant.itemStats
local MM_itemPriceTip = MasterMerchant.itemPriceTip

---
-- @param itemLink @class string
-- @return @class MMstats
function HotepMMX:itemStats(itemLink)
---
-- @local MMstats @class MMstats
  local MMstats = MM_itemStats(MasterMerchant, itemLink)
  
  local itemRec = savedVariables:GetItem(itemLink)
  
  if (#itemRec.sales == 0) then
    savedVariables:DestroyItem(itemLink)
    return MMstats
  end
  
  if (not MMstats.avgPrice) then
    MMstats.avgPrice = itemRec.avgPrice
    MMstats.numDays = itemRec.numDays
    MMstats.numItems = itemRec.numItems
    MMstats.numSales = itemRec.numSales
    return MMstats
  end
  
  
  local maxAge = math.max(MMstats.numDays, itemRec.numDays)
  
  local mmweight = (maxAge - MMstats.numDays) + 1
  local myweight = (maxAge - itemRec.numDays) + 1
  local totalWeight = mmweight + myweight
  
  local price = (MMstats.avgPrice * mmweight) + (itemRec.avgPrice * myweight)
  
  MMstats.avgPrice = price / totalWeight
  MMstats.numDays = tonumber(math.max(MMstats.numDays, itemRec.numDays))
  MMstats.numItems = MMstats.numItems + itemRec.numItems
  MMstats.numSales = MMstats.numSales + itemRec.numSales
  
  return MMstats
end


function HotepMMX:itemPriceTip(itemLink, chatText, clickable)
  
  local line, x, y = MM_itemPriceTip(MasterMerchant, itemLink, chatText, clickable)
  
  local itemRec = savedVariables:GetItem(itemLink)
  
  if (#itemRec.sales == 0) then
    savedVariables:DestroyItem(itemLink)
    return line, x, y
  end
  
  if (line) then
    line = "Hotep\226\132\162 " .. line
  end
  --  	0xE2 0x84 0xA2
  return line, x, y
end




function HotepMMX:RespondPriceCheck(channel, itemLink, handle)
  
  -- pc response:    $$MMX$$price$$sales$$items$$days$$[itemlink]
  
  local foo = function(t)
    HotepMMX:RespondPriceCheck(t.channel, t.itemLink, t.handle)
  end
  
  if (HotepMMX.responding or HotepMMX.asking) then
    Timer:Once(0.25, foo, {channel = channel, itemLink = itemLink, handle = handle})
    return
  end
  
  HotepMMX.responding = true
  
---
-- @local stats @class MMstats
  local stats = HotepMMX:itemStats(itemLink)
  
  if (not stats.avgPrice) then
    HotepMMX.responding = false
    return
  end
  
  
  local resp = {
    "MMX",
    stats.avgPrice,
    stats.numSales,
    stats.numItems,
    stats.numDays,
    itemLink,
  }
  
  local msg = "$$" .. table.concat(resp, "$$")
  
  if (channel == CHAT_CHANNEL_WHISPER) then channel = "wisp" end
  
  local cb = function()
    HotepMMX.responding = false
  end
  
  ChatQueue:New(1, channel, handle, cb, cb, {msg})
end


function HotepMMX:AbsorbPriceResponse(pieces)
  
  -- pc response:    $$MMX$$price$$sales$$items$$days$$[itemlink]
  
  local price = pieces[3]
  local sales = pieces[4]
  local items = pieces[5]
  local days = pieces[6]
  local itemLink = pieces[7]
  
  local itemRec = savedVariables:GetItem(itemLink)
  
  local saleRec = clone(MMXSALE)
  
  saleRec.timestamp = GetTimeStamp()
  saleRec.avgPrice = tonumber(price)
  saleRec.numDays = tonumber(days)
  saleRec.numItems = tonumber(items)
  saleRec.numSales = tonumber(sales)
  
  table.insert(itemRec.sales, saleRec)
  
  HotepMMX:RecalcItem(itemRec)
  
  msgWithName(zo_strformat("Price Data updated for [<<1>>]", itemLink))
end



---
-- @param itemRec @class MMXITEM
function HotepMMX:RecalcItem(itemRec)
  
  
  if (#itemRec.sales == 1) then
    local saleRec = itemRec.sales[1]
    itemRec.avgPrice = saleRec.avgPrice
    itemRec.numDays = saleRec.numDays
    itemRec.numItems = saleRec.numItems
    itemRec.numSales = saleRec.numSales
    return
  end
  
  
  local maxAge = HotepMMX:Oldest(itemRec)
---
-- @local totals @class MMXSALE
  local totals = clone(MMXSALE)
  
  local totalWeight = 0
---
-- @local saleRec @class MMXSALE
  for _,saleRec in pairs(itemRec.sales) do
    local age = math.floor(GetDiffBetweenTimeStamps(GetTimeStamp(), saleRec.timestamp) / ONE_DAY)
    age = age + saleRec.numDays
    
    local weight = (maxAge - age) + 1
    
    totals.avgPrice = totals.avgPrice + (saleRec.avgPrice * weight)
    totalWeight = totalWeight + weight
    
    totals.numDays = math.max(totals.numDays, saleRec.numDays)
    totals.numItems = totals.numItems + saleRec.numItems
    totals.numSales = totals.numSales + saleRec.numSales
  end
  
  totals.avgPrice = totals.avgPrice / totalWeight
  
  itemRec.avgPrice = totals.avgPrice
  itemRec.numDays = totals.numDays
  itemRec.numItems = totals.numItems
  itemRec.numSales = totals.numSales
end


---
-- @param itemRec @class MMXITEM
function HotepMMX:Oldest(itemRec)
  
  local old = 0
  
---
-- @local saleRec @class MMXSALE
  for _,saleRec in pairs(itemRec.sales) do
    
    local age = math.floor(GetDiffBetweenTimeStamps(GetTimeStamp(), saleRec.timestamp) / ONE_DAY)
    age = age + saleRec.numDays
    
    old = math.max(old, age)
  end
  
  return old
end


function HotepMMX:AskPriceCheck(itemLink, stillAsking)
  
  -- ask for pc:     $$MMPC$$[itemlink]
  
  if (HotepMMX.responding or (HotepMMX.asking and not stillAsking)) then return end
  
  HotepMMX.asking = true
  
  local i = HotepMMX.allguilds.i
  local num = HotepMMX.allguilds.num
  
  HotepMMX.allguilds.itemLink = itemLink
  
  local msg = "$$MMPC$$" .. itemLink
  
  local channel = nil
  
  local guildslashes = {"/g1", "/g2", "/g3", "/g4", "/g5"}
  
  if (num) then channel = guildslashes[i] end
  
  local cbdone = function ()
    HotepMMX.asking = false
  end
  
  local cbnext = function ()
    HotepMMX.allguilds.i = HotepMMX.allguilds.i + 1
    HotepMMX:AskPriceCheck(HotepMMX.allguilds.itemLink, true)
  end
  
  
  local cb
  
  if (num) then
    if (i < num) then
      cb = cbnext
    else
      cb = cbdone
    end
  else
    cb = cbdone
  end
  
  ChatQueue:New(1, channel, nil, cb, cb, {msg})
end





function HotepMMX.Slash(extra)
  
  local name = GetItemLinkName(extra)
  
  if ((type(name) ~= "string") or (name == "")) then return false end
  
  HotepMMX.allguilds.num = false
  HotepMMX:AskPriceCheck(extra)
  return true
end

function HotepMMX.SlashAll(extra)
  
  local name = GetItemLinkName(extra)
  
  if ((type(name) ~= "string") or (name == "")) then return false end
  
  local num = GetNumGuilds()
  
  HotepMMX.allguilds.num = num
  HotepMMX.allguilds.i = 1
  HotepMMX:AskPriceCheck(extra)
  return true
end

function HotepMMX.SlashHelp()
  msgWithName("/hotep pc [itemLink] -- price check on current chat")
  msgWithName("/hotep pcguilds [itemLink] -- price check on ALL guild channels")
end




function HotepMMX:CleanupDB()
  
  if (GetDiffBetweenTimeStamps(GetTimeStamp(), Cleanup.last) < ONE_DAY) then
    return
  end
  
  if (Cleanup.mule < 16) then
    Cleanup.mule = Cleanup.mule + 1
  else
    Cleanup.mule = 1
  end
  
  msgWithName("Doing some housekeeping...")
  HotepMMX:CleanupMule(Cleanup.mule)
end

function HotepMMX:CleanupMule(i)
  
  local items = savedVariables:GetMule(i)
  
  local iter = Iterator:New(items, HotepMMX.CleanupItem, items, HotepMMX.CleanupDone)
  
  iter:Next()
end

---
-- @param iter @class Iterator
function HotepMMX.CleanupItem(iter)
  
  local items = iter.params
  local i = iter.index
  
---
-- @local itemRec @class MMXITEM
  local itemRec = items[i]
  
---
-- @local saleRec @class MMXSALE
  for k, saleRec in pairs(itemRec.sales) do
    if (GetDiffBetweenTimeStamps(GetTimeStamp(), saleRec.timestamp) > TOO_OLD) then
      table.remove(itemRec.sales, k)
    end
  end
  
  HotepMMX:RecalcItem(itemRec)
  
  iter:Next()
end

function HotepMMX.CleanupDone()
  msgWithName("Done housekeeping.")
  Cleanup.last = GetTimeStamp()
end




-- ****************************************************************************
--                              event handling
-- ****************************************************************************



function HotepMMX.Initialize()
  
  EVENT_MANAGER:UnregisterForEvent(HotepMMX.name, EVENT_PLAYER_ACTIVATED)
  
  EVENT_MANAGER:RegisterForEvent(HotepMMX.name, EVENT_CHAT_MESSAGE_CHANNEL, HotepMMX.OnChatEvent)
  
  savedVariables:Load()
  
  HotepToolsLib:Init()
  
  
  MasterMerchant.itemStats = HotepMMX.itemStats
  MasterMerchant.itemPriceTip = HotepMMX.itemPriceTip
  
  
  HotepToolsLib.SlashHotep:Register("pc", HotepMMX.Slash)
  HotepToolsLib.SlashHotep:Register("pcguilds", HotepMMX.SlashAll)
  HotepToolsLib.SlashHotep:RegisterHelp(HotepMMX.name, HotepMMX.SlashHelp)
  
  msgWithName("Loaded!")
  HotepMMX.SlashHelp()
  
  HotepMMX:CleanupDB()
end


function HotepMMX.OnChatEvent(eventCode, messageType, fromName, text, isCustomerService)
  
  if (isCustomerService) then return end
  
  if (messageType == CHAT_CHANNEL_WHISPER_SENT) then return end
  
  if ((fromName == GetUnitDisplayName("player")) 
              or (zo_strformat(SI_UNIT_NAME, fromName) == GetUnitName("player"))) then
    return
  end
  
  -- ask for pc:     $$MMPC$$[itemlink]
  -- pc response:    $$MMX$$price$$sales$$items$$days$$[itemlink]
  
  
  local pieces = explode("$$", text)
  
  if (pieces[1] ~= "") then return end
  
  if (pieces[2] == "MMPC") then
    HotepMMX:RespondPriceCheck(messageType, pieces[3], fromName)
    return
  end
  
  if (pieces[2] == "MMX") then
    HotepMMX:AbsorbPriceResponse(pieces)
    return
  end
  
  
end









function HotepMMX.OnAddOnLoaded(event, addonName)
  if addonName == HotepMMX.name then
    math.randomseed(GetTimeStamp())

    EVENT_MANAGER:UnregisterForEvent(HotepMMX.name, EVENT_ADD_ON_LOADED)

    EVENT_MANAGER:RegisterForEvent(HotepMMX.name, EVENT_PLAYER_ACTIVATED, HotepMMX.Initialize)
  end
end


EVENT_MANAGER:RegisterForEvent(HotepMMX.name, EVENT_ADD_ON_LOADED, HotepMMX.OnAddOnLoaded)









